/////////////////////////////////////////////////////////////////////////////
//                                                                         //
// grid.cpp                                                                //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////
//                                                                         //
// Grid control library - Version 1.34 - 11 mar 1995                       //
// Author: H. Picot                                                        //
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

#include "griddll.h"
#pragma hdrstop
#include <ctl3d.h>


// TGrid ----------------------------------------------------------------- //
DEFINE_RESPONSE_TABLE1 (TGrid, TControl)
   EV_WM_CTLCOLOR,
   EV_WM_ERASEBKGND,
   EV_WM_KEYDOWN,
   EV_WM_KILLFOCUS,
   EV_WM_LBUTTONDOWN,
   EV_WM_LBUTTONUP,
   EV_WM_LBUTTONDBLCLK,
   EV_WM_RBUTTONDOWN,
   EV_WM_RBUTTONUP,
   EV_WM_RBUTTONDBLCLK,
   EV_WM_MOUSEMOVE,
   EV_WM_PAINT,
   EV_WM_SETCURSOR,
   EV_WM_SETFOCUS,
   EV_WM_SHOWWINDOW,
   EV_WM_WINDOWPOSCHANGED,
   EV_WM_WINDOWPOSCHANGING,
END_RESPONSE_TABLE;


// ----------------------------------------------------------------------- //
// constructors, destructor, initialization                                //
// ----------------------------------------------------------------------- //
TGrid::TGrid (TWindow *p, int id, int x0, int y0, int w0, int h0,
              TGrid::EnterKey ek, TGrid::TabKey tk, TGrid::LineMarker lm,
              int max, int attrib, TModule *module)
    : TControl  (p, id, "", x0, y0, w0, h0, module)
   {
   Init (max, ek, tk, lm, attrib);
   }


TGrid::TGrid (TWindow *p, int id, TGrid::EnterKey ek, TGrid::TabKey tk,
              TGrid::LineMarker lm, int max, int attrib, TModule *module)
    : TControl  (p, id, module)
   {
   Attr.Style = BS_OWNERDRAW|WS_CHILD|WS_VISIBLE|
                WS_BORDER|WS_TABSTOP|WS_CLIPSIBLINGS;
   Init (max, ek, tk, lm, attrib);
   }


void TGrid::Init (int max, TGrid::EnterKey ek, TGrid::TabKey tk,
                  TGrid::LineMarker lm, int attrib)
   {
   firstColumn = lastColumn = current = 0;
   nControls = hpos = 0;
   logicalWidth = attrWidth = visibleWidth = totalWidth = 0;
   lineHeight = attrHeight = visibleHeight = totalHeight = 0;
   firstLine = activeLine = visibleLines = 0;
   borderX = borderY = gridX = gridY = 0;

   if (max >= 0)
      nLines = max;
   else
      nLines = 0;
   maxLines = max;

   lineColor = TColor::Gray;
   emptyGrid = GetSysColor (COLOR_WINDOW);
   hilitetext = GetSysColor(COLOR_HIGHLIGHTTEXT);
   hilite = GetSysColor(COLOR_HIGHLIGHT);

   firstPaint = true;
   vscrollHidden = false;
   hscrollHidden = false;
   delayFirstFocus = false;
   tempNoHMove = false;
   staticOnly = true;

   enterKey = ek;
   tabKey = tk;
   lineMarker = lm;
   visible = gControl;

   vscroll = (attrib & gVScroll) ? new VScroll (this)     : 0;
   hscroll = (attrib & gHScroll) ? new HScroll (this)     : 0;
   title   = (attrib & gTitle)   ? new Title   (this)     : 0;
   vbar    = (attrib & gVBar)    ? new VerticalBar (this) : 0;

   lineSwapping   = (attrib & gLineSwapOn)      ? true : false;
   centerTitle    = (attrib & gCenterTitle)     ? true : false;
   disableHScroll = (attrib & gDisableHScroll)  ? true : false;
   disableVScroll = (attrib & gDisableVScroll)  ? true : false;
   hideHScroll    = (attrib & gHideHScroll)     ? true : false;
   hideVScroll    = (attrib & gHideVScroll)     ? true : false;
   sizeColumn     = (attrib & gSizeColumn)      ? true : false;
   horzShrink     = (attrib & gHorzShrink)      ? true : false;

   lineSelection = (attrib & gSingleSelection) ? Single : None;
   if (lineSelection == Single)
      {
      if ((attrib & gMultipleSelection) != 0)
         lineSelection = SingleAutoSelect;
      selectedLines = new int[maxsel = nsel = 1];
      selectedLines[0] = -1;
      }
   else if ((attrib & gMultipleSelection) != 0)
      {
      lineSelection = Multiple;
      selectedLines = new int[maxsel = 10];
      nsel = 0;
      }
   else
      selectedLines = 0;

   frame = new Frame (this);

   gridFont = 0;
   titleFont = 0;

   ComboBmp       =
   CheckBoxesBmp  =
   BwccCheckBmp   =
   BwccUncheckBmp =
   Ctl3dCheckBmp  = 0;

   titleLogFont = 0;
   gridLogFont = new LOGFONT;
   memset (gridLogFont, 0, sizeof(LOGFONT));
   strcpy (gridLogFont->lfFaceName, "MS Sans Serif");
   gridLogFont->lfHeight = MulDiv(TScreenDC().GetDeviceCaps (LOGPIXELSY),9,72);
   gridLogFont->lfWeight = FW_BOLD;
   }


TGrid::~TGrid ()
   {
   if (gridLogFont)
      delete gridLogFont;
   if (titleLogFont)
      delete titleLogFont;
   if (firstColumn)
      delete firstColumn;
   if (gridFont)
      delete gridFont;
   if (titleFont)
      delete titleFont;
   if (ComboBmp)
      delete ComboBmp;
   if (CheckBoxesBmp)
      delete CheckBoxesBmp;
   if (BwccCheckBmp)
      delete BwccCheckBmp;
   if (BwccUncheckBmp)
      delete BwccUncheckBmp;
   if (Ctl3dCheckBmp)
      delete Ctl3dCheckBmp;
   if (selectedLines)
      delete [] selectedLines;
   }


void TGrid::PerformCreate (int menuOrId)
   {
   Attr.Style = BS_OWNERDRAW|WS_CHILD|WS_VISIBLE|
                WS_BORDER|WS_TABSTOP|WS_CLIPSIBLINGS;
   TControl::PerformCreate (menuOrId);
   }


char *TGrid::GetClassName()
   {
   return "TGrid";
   }


void TGrid::GetWindowClass(WNDCLASS &wc)
   {
   TWindow::GetWindowClass(wc);
   wc.hbrBackground = (HBRUSH)GetStockObject(LTGRAY_BRUSH);
   }


void TGrid::SetupWindow ()
   {
   TControl::SetupWindow ();
   TApplication *app = GetApplication ();
   if (app && app->Ctl3dEnabled())
      {
      bool FAR PASCAL (*SubclassDlg)(HWND, WORD);
      (FARPROC)SubclassDlg = app->GetCtl3dModule()->GetProcAddress("Ctl3dSubclassDlg");
      if (SubclassDlg)
         SubclassDlg (*this, CTL3D_ALL);
      }
   }


void TGrid::Add (TControl *p, const char *t, Align a)
   {
   if (p)
      {
      Column *ge = new Column (p, lastColumn, this, t);
      lastColumn = ge;
      if (firstColumn == 0)
         firstColumn = ge;

      if (dynamic_cast<TGDateEdit *>(p))  // must be before TGEdit
         {
         dynamic_cast<TGDateEdit *>(p)->grid = this;
         ge->type = gDateEdit;
         staticOnly = false;
         }
      else if (dynamic_cast<TGTimeEdit *>(p))  // must be before TGEdit
         {
         dynamic_cast<TGTimeEdit *>(p)->grid = this;
         ge->type = gTimeEdit;
         staticOnly = false;
         }
      else if (dynamic_cast<TGDoubleEdit *>(p))  // must be before TGEdit
         {
         dynamic_cast<TGDoubleEdit *>(p)->grid = this;
         ge->type = gDoubleEdit;
         staticOnly = false;
         }
      else if (dynamic_cast<TGLongEdit *>(p))  // must be before TGEdit
         {
         dynamic_cast<TGLongEdit *>(p)->grid = this;
         ge->type = gLongEdit;
         staticOnly = false;
         }
      else if (dynamic_cast<TGShortEdit *>(p))  // must be before TGEdit
         {
         dynamic_cast<TGShortEdit *>(p)->grid = this;
         ge->type = gShortEdit;
         staticOnly = false;
         }
      else if (dynamic_cast<TGEdit *>(p))  // must be before TGStatic
         {
         dynamic_cast<TGEdit *>(p)->grid = this;
         ge->type = gEdit;
         staticOnly = false;
         }
      else if (dynamic_cast<TGCheckBox *>(p)) // must be before TGButton
         {
         dynamic_cast<TGCheckBox *>(p)->grid = this;
         ge->type = gCheck;
         staticOnly = false;
         }
      else if (dynamic_cast<TGButton *>(p))
         {
         dynamic_cast<TGButton *>(p)->grid = this;
         ge->type = gButton;
         staticOnly = false;
         }
      else if (dynamic_cast<TGComboBox *>(p))
         {
         dynamic_cast<TGComboBox *>(p)->grid = this;
         ge->type = gCombo;
         staticOnly = false;
         }
      else if (dynamic_cast<TGStatic *>(p))
         {
         dynamic_cast<TGStatic *>(p)->grid = this;
         ge->type = gStatic;
         }

      nControls++;

      if (ge->type == gButton)
         ge->align = Center;
      else
         ge->align = a;
      }
   }



// ----------------------------------------------------------------------- //
// Locating elements (i.e. columns)                                        //
// ----------------------------------------------------------------------- //
TGrid::Column * TGrid::FindColumn (TPoint &p) const
   {
   for (Column *ctrl = firstColumn; ctrl; ctrl = ctrl->next)
      {
      if (TRect (ctrl->left, 0, ctrl->left+ctrl->width, visibleHeight).Contains (p))
         return ctrl;
      }
   return 0;
   }


TGrid::Column *TGrid::FindColumn (TControl *c) const
   {
   for (Column *ctrl = firstColumn; ctrl; ctrl = ctrl->next)
      if (ctrl->control == c)
         return ctrl;

   return 0;
   }


TGrid::Column *TGrid::FindColumn (HWND hwnd) const
   {
   for (Column *ctrl = firstColumn; ctrl; ctrl = ctrl->next)
      if (ctrl->control->HWindow == hwnd)
         return ctrl;

   return 0;
   }


TGrid::Column *TGrid::FindColumn (int id) const
   {
   for (Column *ctrl = firstColumn; ctrl; ctrl = ctrl->next)
      if (ctrl->control->GetId() == id)
         return ctrl;

   return 0;
   }


int TGrid::GetDataLine (int y)
   {
   int y0 = 0;
   for (int i = 0; i < visibleLines; i++, y0 += lineHeight)
      {
      if (y >= y0 && y < y0+lineHeight)
         return i + firstLine; 
      }

   return 0;
   }


void TGrid::FindNearestUnlockedCell (TGrid::Column *e)
   {
   Column *prev = e->previous;
   Column *next = e->next;
   while (prev || next)
      {
      if (prev)
         {
         if (prev->SetFocus ())
            {
            current = prev;
            return;
            }
         prev = prev->previous;
         }
      if (next)
         {
         if (next->SetFocus ())
            {
            current = next;
            return;
            }
         next = next->next;
         }
      }
   }


// ----------------------------------------------------------------------- //
// Line functions                                                          //
// ----------------------------------------------------------------------- //
void TGrid::SetLineHeight (int h)
   {
   if (h == 0)
      return;
   SetVisibleLines (visibleHeight / (lineHeight=h+1));
   }


void TGrid::SetNLines (int n)
   {
   nLines = n;

   if (GetCurrentLine() >= n)
      Go (n-1);

   if (vscroll && vscroll->IsWindow())
      vscroll->SetRange (0, nLines-1);
   }


void TGrid::SetVisibleLines (int nl)
   {
   visibleLines = nl;
   if (vscroll)
      vscroll->PageMagnitude = visibleLines;
   }


void TGrid::SetMaxLines (int m)
   {
   if (m < 0)
      nLines = maxLines;
   else
      nLines = m;
   if (nLines < 0)
      nLines = 0;

   if (GetCurrentLine() >= nLines)
      Go (nLines-1);

   maxLines = m;
   if (vscroll && vscroll->IsWindow())
      vscroll->SetRange (0, maxLines < 0 ? nLines-1 : maxLines-1);
   }


void TGrid::ResetPosition ()
   {
   Go (0);
   GotoFirstColumn ();
   }


void TGrid::SwapLines (int l1, int l2)
   {
   if (l1 == l2 || l1 < 0 || l2 < 0)
      return;
   if (maxLines >= 0 && (l1 >= maxLines || l2 >= maxLines))
      return;

   frame->SetFocus ();

   for (Column *e = firstColumn; e; e = e->next)
      {
      int id = e->control->GetId ();
      switch (e->type)
         {
         // strings (variable size)
         case gStatic:
         case gEdit:
         case gCombo:
            {
            string t1 = static_cast<const char *>(TransferItem (l1, id));
            string t2 = static_cast<const char *>(TransferItem (l2, id));
            TransferItem (l1, id, t2.c_str());
            TransferItem (l2, id, t1.c_str());
            break;
            }

         // other types (fixed size)
         default:
            {
            int l = e->control->Transfer (0, tdSizeData);
            char *t1 = new char[l];
            char *t2 = new char[l];
            memcpy (t1, TransferItem (l1, id), l);
            memcpy (t2, TransferItem (l2, id), l);
            TransferItem (l1, id, t2);
            TransferItem (l2, id, t1);
            delete [] t1;
            delete [] t2;
            break;
            }
         }
      }

   int line = l1 - firstLine;
   if (line >= 0 && line < visibleLines)
      {
      int y = line*lineHeight;
      InvalidateRect (TRect (0, y, visibleWidth-2, y+lineHeight-1), false);
      }

   line = l2 - firstLine;
   if (line >= 0 && line < visibleLines)
      {
      int y = line*lineHeight;
      InvalidateRect (TRect (0, y, visibleWidth-2, y+lineHeight-1), false);
      }

   if (current && !current->SetFocus ())
      FindNearestUnlockedCell (current);
   }


void TGrid::Move (int delta, bool showControl)
   {
   Go (GetCurrentLine()+delta, showControl);
   }


void TGrid::Go (int pos, bool showControl)
   {
   // verifying pos ------------------------------------------------------ //
   if (pos >= nLines)
      {
      if (maxLines > 0)
         pos = nLines-1;
      else
         {
         // Ask permission to append; -------------------------------------//
         // AppendLine may change pos (passed by reference) ---------------//
         if (!AppendNewLine (pos))
            return;

         SetNLines (pos+1);
         }
      }

   if (nLines == 0)
      return;

   if (pos < 0)
      pos = 0;

   // if no change, abort function ----------------------------------------//
   int oldCurrent = GetCurrentLine();
   if (pos == oldCurrent)
      {
      if (lineSelection == SingleAutoSelect)
         SelectLine (pos);
      return;
      }

   // check current cell --------------------------------------------------//
   if (!CanKillFocus (current))
      return;

   // check data ----------------------------------------------------------//
   if (!ValidateLine (oldCurrent))
      return;

   // set focus to frame, while changing current line ---------------------//
   if (frame->SetFocus () == NULL)
      return;     // WM_KILLFOCUS failed

   // remove markers ------------------------------------------------------//
   if (vbar && vbar->IsWindow())
      vbar->ShowText (activeLine);

// New section March 3, 1995 -------------------------------------------------
   // flush controls !!! --------------------------------------------------//
   for (Column *col = firstColumn; col; col = col->next)
      {
      TStatic *s = dynamic_cast<TStatic *>(col->control);
      if (s)
         s->SetText ("");
      else
         {
         TCheckBox *c = dynamic_cast<TCheckBox *>(col->control);
         if (c)
            c->SetCheck (BF_GRAYED);
         else
            {
            TComboBox *c = dynamic_cast<TComboBox *>(col->control);
            if (c)
               c->SetSelIndex (-1);
            else
               {
               TButton *b = dynamic_cast<TButton *>(col->control);
               if (b)
                  b->SetCaption ("");
               }
            }
         }
      }
// End of new section March 3, 1995 ------------------------------------------

   // compute new position ------------------------------------------------//
   int delta = pos - oldCurrent;
   int oldFirstLine = firstLine;
   if (delta + activeLine >= 0 && delta + activeLine < visibleLines)
      activeLine += delta;

   else
      {
      firstLine += delta;
      if (firstLine+visibleLines > nLines && maxLines > 0)
         firstLine = nLines-visibleLines;
      if (firstLine < 0)
         firstLine = 0;
      activeLine = pos - firstLine;
      if (activeLine < 0)
         activeLine = 0;
      else if (activeLine >= visibleLines)
         activeLine = visibleLines-1;
      }

   // redraw grid or scroll -----------------------------------------------//
   if (firstLine != oldFirstLine)
      {                                    
      int scroll = oldFirstLine-firstLine;
      TRect clip (1, 1, visibleWidth-3, visibleHeight-3);
      if (scroll >= visibleLines || scroll <= -visibleLines)
         InvalidateRect (clip, false);
      else
         {
         int s = scroll*lineHeight;
         ScrollWindowEx (0, s, 0, &clip, 0,0, SW_INVALIDATE|SW_SCROLLCHILDREN);
         }

      if (vbar && vbar->IsWindow())
         {
         TRect vbclip (0, 0, vbar?vbar->Attr.W-1:0, vbar?vbar->Attr.H-1:0);
         vbar->InvalidateRect (vbclip, false);
         vbar->UpdateWindow ();
         }
      }

   // transfer data from memory -------------------------------------------//
   int newCurrent = GetCurrentLine ();

   // update scroller -----------------------------------------------------//
   if (vscroll && vscroll->IsWindow())
      vscroll->SetPosition (newCurrent);

   // update vertical bar -------------------------------------------------//
   if (vbar && firstLine == oldFirstLine)
      vbar->ShowMarker (activeLine);

   // reset focus ---------------------------------------------------------//
   if (current && showControl)
      {
      tempNoHMove = true;
      if (!current->SetFocus ())
         FindNearestUnlockedCell (current);
      tempNoHMove = false;
      }

   // Notify user ---------------------------------------------------------//
   if (firstLine != oldFirstLine)
      ScrollData (firstLine, firstLine+visibleLines-1);

   if (newCurrent != oldCurrent)
      {
      NewRowSelected (newCurrent, oldCurrent);
      if (lineSelection == SingleAutoSelect)
         SelectLine (newCurrent);
      }

   if (vscroll && staticOnly && vscroll->IsWindow())
      Parent->PostMessage (WM_NEXTDLGCTL, WPARAM(vscroll->HWindow), 1);
   }


// ----------------------------------------------------------------------- //
// Line selection functions                                                //
// ----------------------------------------------------------------------- //
static int cmpint (const void *a, const void *b)
   {
   return *(int *)a - *(int *)b;
   }


bool TGrid::SelectLine (int line)
   {
   if (lineSelection == Single || lineSelection == SingleAutoSelect)
      {
      // selection outside range of valid lines
      if (line >= 0 && line >= nLines)
         return false;

      // unselecting currently selected line
      else if (line == selectedLines[0] && line >= 0)
         {
         int old = selectedLines[0];
         selectedLines[0] = -1;
         if (old >= 0)
            {
            RefreshLine (line);
            LineSelected (old, false);
            }
         }

      // selecting a new line and unselecting the old one
      else
         {
         int old = selectedLines[0];
         selectedLines[0] = -1;
         RefreshLine (old);
         if (old >= 0)
            LineSelected (old, false);

         selectedLines[0] = line;
         RefreshLine (line);
         if (line >= 0)
            LineSelected (line, true);
         }

      return true;
      }
   else if (lineSelection == Multiple)
      {
      int *s = static_cast<int *>(bsearch (&line, selectedLines, nsel, sizeof(int), cmpint));
      if (s)
         {
         int i = int(s - selectedLines);
         nsel--;
         memcpy (s, s+1, (nsel-i)*sizeof(int));
         RefreshLine (line);
         LineSelected (line, false);
         }
      else
         {
         selectedLines[nsel++] = line;
         RefreshLine (line);
         LineSelected (line, true);
         if (nsel == maxsel)
            {
            int *p = new int[maxsel+10];
            memcpy (p, selectedLines, maxsel*sizeof(int));
            delete [] selectedLines;
            selectedLines = p;
            maxsel += 10;
            }
         qsort (selectedLines, nsel, sizeof(int), cmpint);
         }
      return true;
      }

   return false;
   }


bool TGrid::LineSelected (int line) const
   {
   if (lineSelection == Single || lineSelection == SingleAutoSelect)
      {
      if (line == selectedLines[0])
         return true;
      }
   else if (lineSelection == Multiple)
      {
      int *s = static_cast<int *>(bsearch (&line, selectedLines, nsel, sizeof(int), cmpint));
      return s ? true : false;
      }
   return false;
   }


int TGrid::CountSelectedLines () const
   {
   if (lineSelection == Single || lineSelection == SingleAutoSelect)
      {
      if (selectedLines[0] >= 0)
         return 1;
      }
   else if (lineSelection == Multiple)
      return nsel;

   return 0;
   }


int TGrid::GetSelectedLine ()   const
   {
   if (lineSelection != None)
      return selectedLines[0];

   return -1;
   }


int TGrid::GetSelectedLines (int *list) const
   {
   if (list == 0)
      return 0;

   if (lineSelection == Single || lineSelection == SingleAutoSelect)
      {
      if (selectedLines[0] >= 0)
         {
         list[0] = selectedLines[0];
         return 1;
         }
      }
   else if (lineSelection == Multiple)
      {
      qsort (selectedLines, nsel, sizeof(int), cmpint);
      memcpy (list, selectedLines, nsel*sizeof(int));
      return nsel;
      }

   return 0;
   }


void TGrid::UnselectAllLines ()
   {
   if (lineSelection == Single || lineSelection == SingleAutoSelect)
      {
      int old = selectedLines[0];
      if (old >= 0)
         {
         selectedLines[0] = -1;
         RefreshLine (old);
         LineSelected (old, false);
         }
      }
   else if (lineSelection == Multiple)
      {
      int old = nsel;
      nsel = 0;
      while (old > 0)
         {
         RefreshLine (selectedLines[--old]);
         LineSelected (selectedLines[old], false);
         }
      }
   }


bool TGrid::ValidateLine (int line)
   {
   // retrieve data from current control
   TControl *c = current ? current->control : 0;
   if (c)
      {
      // check if data has been touched
      TGEdit *ed = dynamic_cast<TGEdit *>(c);
      TGComboBox *cb = dynamic_cast<TGComboBox *>(c);
      TGCheckBox *ck = dynamic_cast<TGCheckBox *>(c);
      bool touched = false;
      if (ed)
         touched = ed->IsModified ();
      else if (ck && ck->GetCheck() != ck->oldstate)
         touched = true;
      else if (cb)
         touched = cb->IsModified ();

      if (touched)
         {
         void *data = 0;
         if (current->type == gCombo)
            data = dynamic_cast<TGComboBox *>(c)->GetData ();
         else
            {
            int l = c->Transfer (0, tdSizeData);
            if (l > 0)
               {
               data = new char[l];
               c->Transfer (data, tdGetData);
               }
            }

         int id = c->GetId ();
         bool rc = DataChanging (line, id, data);
         if (rc)
            {
            TransferItem (line, id, data);
            DataChanged (line, id, data);
            }
         if (data)
            delete [] data;
         if (!rc)
            return false;
         }
      }

   // now, validate line
   return IsValidLine (line);
   }

// ----------------------------------------------------------------------- //
// Keyboard response                                                       //
// ----------------------------------------------------------------------- //
bool TGrid::KeyDown (uint key, int repeat)
   {
   bool ctrl  = (GetKeyState (VK_CONTROL) & 0x8000) ? true : false;
   bool shift = (GetKeyState (VK_SHIFT)   & 0x8000) ? true : false;

   switch (key)
      {
      case VK_ESCAPE:   return VkEscape (ctrl, shift, repeat);
      case VK_INSERT:   return VkInsert (ctrl, shift, repeat);
      case VK_DELETE:   return VkDelete (ctrl, shift, repeat);
      case VK_PRIOR:    return VkPrior  (ctrl, shift, repeat);
      case VK_NEXT:     return VkNext   (ctrl, shift, repeat);
      case VK_UP:       return VkUp     (ctrl, shift, repeat);
      case VK_DOWN:     return VkDown   (ctrl, shift, repeat);
      case VK_RIGHT:    return VkRight  (ctrl, shift, repeat);
      case VK_LEFT:     return VkLeft   (ctrl, shift, repeat);
      case VK_RETURN:   return VkReturn (ctrl, shift, repeat);
      case VK_TAB:      return VkTab    (ctrl, shift, repeat);
      case VK_HOME:     return VkHome   (ctrl, shift, repeat);
      case VK_END:      return VkEnd    (ctrl, shift, repeat);
      }

   return false;
   }


bool TGrid::VkEscape (bool ctrl, bool shift, int)
   {
   if (ctrl == false && shift == false)
      {
      Parent->Destroy (IDCANCEL);
      return true;
      }

   return false;
   }


bool TGrid::VkEnd (bool ctrl, bool, int)
   {
   if (ctrl)
      {
      if (!CanKillFocus (current))
         return true;

      GotoLastColumn ();
      return true;
      }

   return false;
   }


bool TGrid::VkHome (bool ctrl, bool, int)
   {
   if (ctrl)
      {
      if (!CanKillFocus (current))
         return true;

      GotoFirstColumn ();
      return true;
      }

   return false;
   }


bool TGrid::VkRight (bool, bool, int repeat)
   {
   for (int i = 0; i < repeat; i++)
      if (!GotoNextColumn ())
         break;
   return true;
   }


bool TGrid::VkLeft (bool, bool, int repeat)
   {
   for (int i = 0; i < repeat; i++)
      if (!GotoPreviousColumn ())
         break;
   return true;
   }


bool TGrid::VkUp (bool ctrl, bool, int repeat)
   {
   if (ctrl)
      {
      if (lineSwapping == false)
         return false;
      for (int i = 0; i < repeat; i++)
         {
         int l = firstLine+activeLine;
         if (l <= 0)
            break;
         SwapLines (l-1, l);
         Move (-1);
         }
      }
   else
      Move (-repeat);
   return true;
   }


bool TGrid::VkDown (bool ctrl, bool, int repeat)
   {
   if (ctrl)
      {
      if (lineSwapping == false)
         return false;
      for (int i = 0; i < repeat; i++)
         {
         int l = firstLine+activeLine;
         if (l >= nLines-1)
            break;
         SwapLines (l, l+1);
         Move (1);
         }
      }
   else
      Move (repeat);
   return true;
   }


bool TGrid::VkNext (bool ctrl, bool, int repeat)
   {
   if (ctrl)
      {
      Go (nLines-visibleLines-1, false);
      Go (nLines-1);
      }
   else
      Move (visibleLines*repeat);
   return true;
   }


bool TGrid::VkPrior (bool ctrl, bool, int repeat)
   {
   if (ctrl)
      Go (0);
   else
      Move (-visibleLines*repeat);
   return true;
   }


bool TGrid::VkReturn (bool ctrl, bool shift, int repeat)
   {
   if (ctrl)
      return false;

   else if (shift && lineSelection != None)
      return SelectLine (GetCurrentLine ());

   switch (enterKey)
      {
      case NextLine:
         {
         if (!CanKillFocus (current))
            return true;

         Move (repeat, false);
         GotoFirstColumn ();
         return true;
         }

      case DoNothing:
         return true;

      case EndDialog:
         return false;

      case EnterNextColumn:
         {
         if (!CanKillFocus (current))
            return true;

         for (int i = 0; i < repeat; i++)
            {
            if (!GotoNextColumn ())
               {
               Move (1, false);
               GotoFirstColumn ();
               }
            }
         return true;
         }
      }
   return false;
   }


bool TGrid::VkTab (bool ctrl, bool shift, int repeat)
   {
   if (ctrl || tabKey == NextControl)
      {
      if (ValidateLine(activeLine+firstLine))
         {
         HWND hwnd = Parent->GetNextDlgTabItem (HWindow, shift);
         if (hwnd)
            {
            if (current)
               current->HideControl ();
            Parent->PostMessage (WM_NEXTDLGCTL, shift);
            }
         }
      return true;
      }
   else
      {
      if (!CanKillFocus (current))
         return true;

      if (tabKey == AutoNextLine)
         {
         if (shift)
            {
            for (int i = 0; i < repeat; i++)
               {
               if (!GotoPreviousColumn ())
                  {
                  Move (-1, false);
                  GotoLastColumn ();
                  }
               }
            }
         else
            {
            for (int i = 0; i < repeat; i++)
               {
               if (!GotoNextColumn ())
                  {
                  Move (1, false);
                  GotoFirstColumn ();
                  }
               }
            }
         return true;
         }
      else
         {
         if (shift)
            return VkLeft (ctrl, shift, repeat);
         else
            return VkRight (ctrl, shift, repeat);
         }
      }
   }


bool TGrid::VkDelete (bool ctrl, bool, int)
   {
   if (ctrl)
      {
      frame->SetFocus ();
      if (DeleteLine (activeLine+firstLine))
         Invalidate ();
      if (current && !current->SetFocus ())
         FindNearestUnlockedCell (current);
      return true;
      }
   return false;
   }


bool TGrid::VkInsert (bool ctrl, bool, int repeat)
   {
   if (ctrl)
      {
      if (!CanKillFocus (current) || !ValidateLine (activeLine+firstLine))
         return true;
      frame->SetFocus ();
      for (int i = 0; i < repeat; i++)
         {
         if (!InsertLine (activeLine+firstLine))
            return true;
         }
      Invalidate ();
      if (current && !current->SetFocus ())
         FindNearestUnlockedCell (current);
      return true;
      }
   return false;
   }

// ----------------------------------------------------------------------- //
// Handling columns                                                        //
// ----------------------------------------------------------------------- //
bool TGrid::GotoPreviousColumn ()
   {
   if (current)
      current->control->SendMessage (WM_KILLFOCUS);

   Column *e = current;
   if (e == 0)
      e = lastColumn;
   if (e)
      e = e->previous;
   for (; e; e = e->previous)
      {
      if (e->SetFocus ())
         return true;
      }
   if (vscroll && staticOnly && vscroll->IsWindow())
      Parent->PostMessage (WM_NEXTDLGCTL, WPARAM(vscroll->HWindow), 1);
   return false;
   }


bool TGrid::GotoNextColumn ()
   {
   if (current)
      current->control->SendMessage (WM_KILLFOCUS);

   Column *e = current;
   if (e == 0)
      e = firstColumn;
   if (e)
      e = e->next;
   for (; e; e = e->next)
      {
      if (e->SetFocus ())
         return true;
      }
   if (vscroll && staticOnly && vscroll->IsWindow())
      Parent->PostMessage (WM_NEXTDLGCTL, WPARAM(vscroll->HWindow), 1);
   return false;
   }


void TGrid::GotoLastColumn ()
   {
   if (current)
      current->control->SendMessage (WM_KILLFOCUS);

   if (current)
      current->HideControl ();
   current = lastColumn;
   if (current && !current->SetFocus ())
      GotoPreviousColumn ();
   }


void TGrid::GotoFirstColumn ()
   {
   if (current)
      current->control->SendMessage (WM_KILLFOCUS);

   if (current)
      current->HideControl ();
   current = firstColumn;
   if (current && !current->SetFocus ())
      GotoNextColumn ();
   }


void TGrid::SetHPos (int newpos)
   {
   MoveHPos (newpos - hpos);
   }


void TGrid::MoveHPos (int delta)
   {
   int pos = hpos + delta;
   if (pos < 0)
      pos = 0;
   else if (visibleWidth >= logicalWidth)
      pos = 0;
   else if (lastColumn->left+lastColumn->width-delta < visibleWidth)
      pos = logicalWidth-visibleWidth;
   delta = pos - hpos;

   for (Column *ctrl = firstColumn; ctrl; ctrl = ctrl->next)
      ctrl->left -= delta;

   TRect clip (1, 1, visibleWidth-3, visibleHeight-3);
   ScrollWindowEx (-delta, 0, 0, &clip, 0,0, SW_INVALIDATE|SW_SCROLLCHILDREN);
   UpdateWindow ();

   if (title && title->IsWindow())
      {
      title->Invalidate ();
      title->UpdateWindow ();
      }

   if (hscroll && hscroll->IsWindow())
      hscroll->SetPosition (hpos = pos);
   }


void TGrid::SetColumnWidth (int id, int width)
   {
   if (firstPaint)
      return;

   Column *e = FindColumn (id);
   int delta = width - e->width;
   if (delta)
      ChangeColWidth (e, delta);
   }


void TGrid::ChangeColWidth (Column *e, int delta)
   {
   if (delta < 0 && delta < -(e->width-8)) // min width is 8 pixels
      delta = -(e->width-8);

   e->width += delta;
   for (Column *col = e->next; col; col = col->next)
      col->left += delta;
   logicalWidth += delta;

   CheckHScroll ();

   if (IsWindow ())
      {
      TRect clip (e->left+e->width-1, 0, visibleWidth-2, visibleHeight-2);
      ScrollWindowEx (delta, 0, 0, &clip, 0,0, SW_INVALIDATE|SW_SCROLLCHILDREN);
      InvalidateRect (TRect (e->left, 0, e->left+e->width, visibleHeight), true);

      if (title && title->IsWindow())
         {
         clip.bottom = title->Attr.H-2;
         title->ScrollWindowEx (delta, 0, 0, &clip, 0,0, SW_INVALIDATE);
         title->InvalidateRect (TRect (e->left, 0, e->left+e->width,
                                title->Attr.H), true);
         title->UpdateWindow ();
         }
      UpdateWindow ();
      }
   }


void TGrid::Remove (int id, bool destroy)
   {
   Column *e = FindColumn (id);
   if (e)
      Remove (e, destroy);
   }


void TGrid::Remove (Column *e, bool destroy)
   {
   if (e == 0)
      return;

   int w = e->width;
   for (Column *col = e->next; col; col = col->next)
      col->left -= w;
   logicalWidth -= w;

   CheckHScroll ();
   InvalidateRect (TRect (e->left, 0, visibleWidth, visibleHeight), false);

   if (title && title->IsWindow())
      title->InvalidateRect (TRect (e->left, 0, visibleWidth, title->Attr.H),
                             false);

   if (current == e)
      {
      if (e->next)
         current = e->next;
      else if (e->previous)
         current = e->previous;
      else
         current = 0;
      }

   e->next->previous = e->previous;
   e->previous->next = e->next;
   if (lastColumn == e)
      lastColumn = e->previous;
   if (firstColumn == e)
      firstColumn = e->next;

   if (destroy)
      delete e->control;

   if (e->type != gStatic)
      {
      staticOnly = true;
      for (Column *c = firstColumn; c; c = c->next)
         if (e->type != gStatic)
            {
            staticOnly = false;
            break;
            }
      }

   e->next = e->previous = 0;
   delete e;
   }


void TGrid::CheckHScroll ()
   {
   // check if horiz scroll has to be shown/hidden --------------------------
   if (hscroll && hscroll->IsWindow())
      {
      bool scrollable = ((logicalWidth > visibleWidth) ? true : false);
      if (hideHScroll)
         {
         bool bVisible = (hscroll->IsWindowVisible () ? true : false);
         if ((bVisible ^ scrollable) != 0)
            SetWindowPos(0,0,0,attrWidth,attrHeight,SWP_NOZORDER|SWP_NOMOVE);
         }
      else if (disableHScroll)
         {
         bool enabled = (hscroll->IsWindowEnabled () ? true : false);
         if ((enabled ^ scrollable) != 0)
            hscroll->EnableWindow (logicalWidth > visibleWidth);
         }
      }

   SetHScrollRange ();
   }


// ----------------------------------------------------------------------- //
// Colors                                                                  //
// ----------------------------------------------------------------------- //
HBRUSH TGrid::EvCtlColor(HDC hDC, HWND hWndChild, uint ctlType)
   {
   Column *e = FindColumn (hWndChild);
   TWindow *wptr = e->control;
   TColor bg, fg;
   HBRUSH brush;

   if (dynamic_cast<TGButton *>(wptr) == 0 && LineSelected (GetCurrentLine()))
      {
      static HBRUSH brush = 0;
      TDC dc(hDC);
      dc.SetTextColor (hilitetext);
      dc.SetBkColor (hilite);
      brush = TBrush (hilite);
      return brush;
      }

   bool customColors = CellColor (GetCurrentLine(), wptr->GetId(), true, bg, fg, brush);
   if (customColors)
      {
      TDC dc(hDC);
      dc.SetTextColor (fg);
      dc.SetBkColor (bg);
      return brush;
      }

   if (dynamic_cast<TGEdit *>(wptr))
      return dynamic_cast<TGEdit *>(wptr)->CtlColors (hDC);

   if (dynamic_cast<TGComboBox *>(wptr))
      return dynamic_cast<TGComboBox *>(wptr)->CtlColors (hDC);

   if (dynamic_cast<TGStatic *>(wptr))
      return dynamic_cast<TGStatic *>(wptr)->CtlColors (hDC);

   if (dynamic_cast<TGCheckBox *>(wptr))
      return dynamic_cast<TGCheckBox *>(wptr)->CtlColors (hDC);

   TApplication *app = GetApplication ();
   if (app->BWCCEnabled () || app->Ctl3dEnabled())
      {
      char buffer[32];
      ::GetClassName (hWndChild, buffer, 32);
      if (stricmp (buffer, "Static") == 0)
         {
         TDC (hDC).SetBkColor (TColor::LtGray);
         return HBRUSH(GetStockObject (LTGRAY_BRUSH));
         }
      }

   if (app->Ctl3dEnabled())
      {
      HBRUSH WINAPI(*CtlColorEx)(uint, WPARAM, LPARAM);
      (FARPROC)CtlColorEx = app->GetCtl3dModule()->GetProcAddress("Ctl3dCtlColorEx");
      HBRUSH hb = 0;
      if (CtlColorEx)
      #if defined(__WIN32__)
        hb = CtlColorEx(WM_CTLCOLORMSGBOX+ctlType, WPARAM(hDC), LPARAM(hWndChild));
      #else
        hb = CtlColorEx(WM_CTLCOLOR, WPARAM(hDC), MAKELPARAM(hWndChild, ctlType));
      #endif
      if (hb != 0)
         return hb;
      }

   return TWindow::EvCtlColor(hDC, hWndChild, ctlType);
   }


// ----------------------------------------------------------------------- //
// Switching focus                                                         //
// ----------------------------------------------------------------------- //
void TGrid::EvSetFocus (HWND)
   {
   if (firstPaint)
      {                  
      current = 0;
      delayFirstFocus = true;
      return;
      }

   if (current == 0)
      GotoFirstColumn ();
   else
      current->SetFocus ();

   RefreshLine (GetCurrentLine ());

   if (vscroll && staticOnly && vscroll->IsWindow())
      Parent->PostMessage (WM_NEXTDLGCTL, WPARAM(vscroll->HWindow), 1);
   }


void TGrid::EvKillFocus (HWND hWndGetFocus)
   {
   if (current && CanKillFocus (current) && current->control->HWindow != hWndGetFocus)
      current->HideControl ();
   }


void TGrid::SetFocusOnChild (TControl *child, HWND /*loosefocus*/)
   {
   if (firstPaint)
      return;

   Column *e = FindColumn (child);
   if (e == 0)
      return;

   int delta = 0;

   if (e->left < 0)
      delta = e->left;
   else if (e->left+e->width > visibleWidth)
      delta = e->left+e->width - visibleWidth;
   if (delta && !tempNoHMove)
      MoveHPos (delta);

   int id = child->GetId ();
   int line = GetCurrentLine ();
   void *data = const_cast<void *>(TransferItem (line, id));
   if (e->type == gCombo)
      dynamic_cast<TGComboBox *>(child)->SetData (data);
   else
      child->Transfer (data, tdSetData);
   e->ShowControl ();
   }


void TGrid::KillFocusFromChild (TControl *c, HWND /*getfocus*/, bool okToChange)
   {
   if (current)
      {
      Column *ctrl = FindColumn (c);
      if (ctrl)
         {
         void *data = 0;
         if (ctrl->type == gCombo)
            data = dynamic_cast<TGComboBox *>(ctrl->control)->GetData ();
         else
            {
            int l = ctrl->control->Transfer (0, tdSizeData);
            if (l > 0)
               {
               data = new char[l];
               ctrl->control->Transfer (data, tdGetData);
               }
            }
         if (okToChange)
            {
            int l = GetCurrentLine ();
            int id = c->GetId ();
            TransferItem (l, id, data);
            DataChanged (l, id, data);
            }
         if (data)
            delete [] data;
         }
      current->HideControl ();
      UpdateWindow ();
      }
   }


bool TGrid::CanKillFocus (Column *e)
   {
   if (e)
      return e->CanKillFocus ();
   else
      return true;
   }

// ----------------------------------------------------------------------- //
// Painting                                                                //
// ----------------------------------------------------------------------- //
void TGrid::EvPaint ()
   {
   TPaintDC dc(*this);
   Paint(dc, dc.Ps.fErase, *(TRect*)&dc.Ps.rcPaint);
   }


void TGrid::Paint (TDC &dc, bool, TRect &clip)
   {
   // terminate initialization ----------------------------------------------
   if (firstPaint == true)
      {
      firstPaint = false;
      CalcPos ();
      Invalidate ();
      return;
      }

   // clip drawing ----------------------------------------------------------
   dc.SelectClipRgn (TRegion(clip));

   // No column to paint ? --------------------------------------------------
   if (firstColumn == 0)
      {
      dc.FillRect (clip, TBrush (emptyGrid));
      return;
      }

   // draw lines ------------------------------------------------------------
   int y = 0;
   for (int i = 0; i < visibleLines; i++, y += lineHeight)
      {
      int pos = firstLine+i;
      if (maxLines < 0 && pos >= nLines)
         {
         if (!AppendNewLine (pos))
            {
            TRect rc (0, y, visibleWidth, visibleHeight);
            dc.FillRect (rc, TBrush (emptyGrid));
            break;
            }
         SetNLines (pos+1);
         }

      TRect currect (0, y, visibleWidth-1, y+lineHeight-1);
      if (!currect.Touches (clip))
         continue;

      // text ---------------------------------------------------------------
      for (Column *ctrl = firstColumn; ctrl; ctrl = ctrl->next)
         {
         // Select font -----------------------------------------------------
         if (ctrl->font)
            dc.SelectObject (*ctrl->font);
         else
            dc.SelectObject (*gridFont);

         // do not draw elements scrolled out horizontally ------------------
         if (ctrl->left + ctrl->width <= 0)
            continue;
         if (ctrl->left >= visibleWidth)
            continue;

         TControl *c = ctrl->control;
         const void *data = TransferItem (i+firstLine, c->GetId ()); 

         // drawing area ----------------------------------------------------
         TRect rc (ctrl->left, y, ctrl->left+ctrl->width-1, y+lineHeight-1);
         if (rc.Touches (clip))
            {
            Visible v = VisibleCell (i+firstLine, ctrl->control->GetId());
            if (v == gEmpty)
               dc.FillRect (rc, TBrush (emptyGrid));
            else
               {
               switch (ctrl->type)
                  {
                  case gTimeEdit:
                  case gDateEdit:
                  case gShortEdit:
                  case gLongEdit:
                  case gDoubleEdit:
                  case gEdit:
                     dynamic_cast<TGEdit *>(c)->DrawCell (dc, rc,
                                                             data, ctrl->align);
                     break;

                  case gStatic:
                     dynamic_cast<TGStatic *>(c)->DrawCell (dc, rc,
                                                             data, ctrl->align);
                     break;

                  case gCombo:
                     dynamic_cast<TGComboBox *>(c)->DrawCell (dc, rc, data);
                     break;

                  case gCheck:
                     dynamic_cast<TGCheckBox *>(c)->DrawCell (dc, rc,
                                                             data, ctrl->align);
                     break;

                  case gButton:
                     dynamic_cast<TGButton *>(c)->DrawCell (dc, rc, data);
                     break;
                  }
               }
            }

         dc.RestoreFont ();
         }
      }

	// draw grid -------------------------------------------------------------
   if (logicalWidth <= visibleWidth)
      {
      TRect rc (logicalWidth, 0, visibleWidth, visibleHeight);
      dc.FillRect (rc, TBrush(emptyGrid));
	   dc.SelectObject (TPen (GetSysColor (COLOR_WINDOWFRAME)));
      int x = logicalWidth-2;
      dc.MoveTo (x, 0);
      dc.LineTo (x, visibleHeight-2);
      dc.RestorePen ();
      }
   clip.Inflate (1);
	dc.SelectClipRgn (TRegion(clip));
	dc.SelectObject (TPen (lineColor));
   y = lineHeight-1;
   for (i = 1; i < visibleLines; i++, y += lineHeight)
      {
      dc.MoveTo (0, y);
      dc.LineTo (visibleWidth-2, y);
      }

   for (Column *ctrl = firstColumn->next; ctrl; ctrl = ctrl->next)
      {
      int x = ctrl->left-1;
      if (x >= 0 && x < visibleWidth)
         {
         dc.MoveTo (x, 0);
         dc.LineTo (x, visibleHeight-2);
         }
      }

   TRect frame = GetClientRect ();
   dc.FrameRect (frame, TBrush (TColor::Black));

   dc.RestorePen ();
   }


void TGrid::RefreshLine (int line)
   {
   if (line < firstLine || line >= firstLine+visibleLines)
      return;

   int gridline = line - firstLine;
   int y = lineHeight * gridline;
   TRect clip (0, y, visibleWidth-2, y+lineHeight);
   InvalidateRect (clip, false);

   // update vbar -----------------------------------------------------------
   RefreshItem (line, 0);
   }


void TGrid::RefreshItem (int line, int id)
   {
   if (line < firstLine || line >= firstLine+visibleLines)
      return;

   int gridline = line - firstLine;

   if (gridline == activeLine)
      {
      Column *e = FindColumn (id);
      if (e)
         {
         void *data = const_cast<void *>(TransferItem (line, id));
         if (e->type == gCombo)
            dynamic_cast<TGComboBox *>(e->control)->SetData (data);
         else
            e->control->Transfer (data, tdSetData);
         }
      }

   int y = lineHeight * gridline;
   if (id != 0)   // update grid data
      {
      Column *e = FindColumn (id);
      TRect clip (e->left, y, e->left+e->width, y+lineHeight);
      if (clip.right < 0 || clip.left >= visibleWidth)
         return;
      if (clip.left < 0)
         clip.left = 0;
      if (clip.right >= visibleWidth-1)
         clip.right = visibleWidth-2;
      if (!clip.IsEmpty ())
         InvalidateRect (clip, false);
      }
   else if (vbar && vbar->IsWindow())  // update vbar
      {
      TRect clip (0, y, vbar->Attr.W-2, y+lineHeight);
      vbar->InvalidateRect (clip, false);
      }
   }


void TGrid::Refresh ()
   {
   for (int i = 0; i < visibleLines; i++)
      RefreshLine (firstLine+i);
   }


void TGrid::EvShowWindow (bool show, uint status)
   {
   TControl::EvShowWindow (show, status);

   int command = show ? SW_SHOW : SW_HIDE;
   if (frame && frame->IsWindow ())
      frame->ShowWindow (command);
   if (vbar && vbar->IsWindow ())
      vbar->ShowWindow (command);
   if (title && title->IsWindow ())
      title->ShowWindow (command);
   if (vscroll && vscroll->IsWindow () && !vscrollHidden)
      vscroll->ShowWindow (command);
   if (hscroll && hscroll->IsWindow () && !hscrollHidden)
      hscroll->ShowWindow (command);
   }

// ----------------------------------------------------------------------- //
// Changing grid size                                                      //
// ----------------------------------------------------------------------- //
void TGrid::EvWindowPosChanging (WINDOWPOS far &wpos)
   {
   ChangeWindowPos (wpos.x, wpos.y, wpos.cx, wpos.cy, wpos.flags);
   TControl::EvWindowPosChanging (wpos);
   }


void TGrid::ChangeWindowPos (int &x, int &y, int &w, int &h, uint flags)
   {
   // do nothing if initialization is not finished --------------------------
   if (firstPaint)
      return;

   int oldFirstLine = firstLine;
   int oldVisibleLines = visibleLines;

   bool Move = ((flags & SWP_NOMOVE) == 0 ? true : false);
   bool Size = ((flags & SWP_NOSIZE) == 0 ? true : false);

   const int cxvscroll = GetSystemMetrics(SM_CXVSCROLL)-1;
   const int cyhscroll = GetSystemMetrics(SM_CYHSCROLL)-1;

   if (flags & SWP_NOMOVE)
      {
      x = borderX;
      y = borderY;
      }
   else
      {
      borderX = x;
      borderY = y;
      }

   gridX = x;
   gridY = y;

   if (flags & SWP_NOSIZE)
      {
      visibleWidth = totalWidth;
      visibleHeight = totalHeight;
      }
   else
      {
      visibleWidth = attrWidth = w;
      visibleHeight = attrHeight = h;
      }

   if (vbar)
      {
      int &vw = vbar->Attr.W;
      vw = vbar->CalcWidth ();
      gridX += (vw -1);
      visibleWidth -= (vw -1);
      }

   if (title)
      {
      int &th = title->Attr.H;
      th = title->CalcHeight ();
      gridY += (th -1);
      visibleHeight -= (th -1);
      }

   if (vscroll)
      visibleWidth -= cxvscroll;

   if (hscroll)
      visibleHeight -= cyhscroll;

   if (lastColumn->left+lastColumn->width < visibleWidth && firstColumn->left <= 0)
      MoveHPos (lastColumn->left+lastColumn->width - visibleWidth);

   if (Size)
      {
      if (hscroll && hscroll->IsWindow())
         {
         if (logicalWidth <= visibleWidth)
            {
            if (disableHScroll)
               hscroll->EnableWindow (false);
            if (hideHScroll)
               {
               if (!hscrollHidden)
                  {
                  hscroll->ShowWindow (SW_HIDE);
                  hscrollHidden = true;
                  }
               visibleHeight += cyhscroll;
               }
            }
         else
            {
            if (!hscroll->IsWindowVisible ())
               {
               hscroll->ShowWindow (SW_SHOW);
               hscrollHidden = false;
               }
            hscroll->EnableWindow (true);
            }
         }
      }

   if (Move || Size)
      {
      uint flag = SWP_NOZORDER;

      if (!Size)
         flag |= SWP_NOSIZE;
      else if (!Move)
         flag |= SWP_NOMOVE;

      // adjust to integral line height -------------------------------------
      if (h && lineHeight)
         {
         int old = visibleLines;
         int curline = firstLine+activeLine;
         SetVisibleLines (visibleHeight / lineHeight);
         if (maxLines > 0 && visibleLines >= maxLines)
            {
            if (vscroll && vscroll->IsWindow())
               {
               if (vscroll->IsWindowVisible())
                  {
                  if (disableVScroll)
                     vscroll->EnableWindow (false);
                  if (hideVScroll)
                     {
                     vscroll->ShowWindow (SW_HIDE);
                     vscrollHidden = true;
                     }
                  }
               if (hideVScroll)
                  visibleWidth += cxvscroll;
               }
            SetVisibleLines (maxLines);
            }
         else if (vscroll && vscroll->IsWindow())
            {
            if (!vscroll->IsWindowVisible ())
               {
               vscroll->ShowWindow (SW_SHOW);
               vscrollHidden = false;
               }
            vscroll->EnableWindow (true);
            }

         if (visibleLines <= 0)
            SetVisibleLines (1);

         if (visibleLines != old)
            {
            Go (0);
            if (curline)
               Go (curline);
            }

         visibleHeight = visibleLines * lineHeight + 1;
         }

      // adjust related controls -----------------------------------------------
      bool vs = (vscroll && vscroll->IsWindow() && vscroll->IsWindowVisible ()) ? true : false;
      bool hs = (hscroll && hscroll->IsWindow() && hscroll->IsWindowVisible ()) ? true : false;
      const int left = vbar ? vbar->Attr.W-1 : 0;
      const int top = title ? title->Attr.H-1 : 0;
      const int right = vs ? cxvscroll : 0;
      const int bottom = hs ? cyhscroll : 0;

      if (horzShrink && visibleWidth > logicalWidth)
         visibleWidth = logicalWidth;

      totalWidth = visibleWidth + left + right;
      totalHeight = visibleHeight + top + bottom;
      if (visibleWidth < 0)
         visibleWidth = 0;
      if (totalHeight < 0)
         totalHeight = 0;

      flag &= (~SWP_NOZORDER); // borders are after grid in z-order ------------

      if (frame)
         frame->SetWindowPos (HWindow, borderX, borderY, totalWidth,
                              totalHeight, flag);

      if (title && title->IsWindow())
         title->SetWindowPos (HWindow, gridX, borderY, visibleWidth,
                              title->Attr.H, flag);

      if (vbar && vbar->IsWindow())
         vbar->SetWindowPos (HWindow, borderX, gridY, vbar->Attr.W,
                             visibleHeight, flag);

      flag &= (~SWP_NOMOVE);   // vscroll and hscroll are always moved ---------

      if (vs)
         vscroll->SetWindowPos (HWindow, gridX+visibleWidth, gridY, right,
                                visibleHeight, flag);

      if (hs)
         {
         hscroll->SetWindowPos (HWindow, gridX, gridY+visibleHeight,
                                visibleWidth, bottom, flag);
         SetHScrollRange ();
         }
      }

   x = gridX;
   y = gridY;
   w = visibleWidth;
   h = visibleHeight;

   if (firstLine != oldFirstLine || visibleLines != oldVisibleLines)
      ScrollData (firstLine, firstLine+visibleLines-1);
   }


void TGrid::CalcPos ()
   {
   // create the grid's TFont -----------------------------------------------
   SetFont ();
   if (titleLogFont)
      SetTitleFont ();

   // line height -----------------------------------------------------------
   for (Column *ctrl = firstColumn; ctrl; ctrl = ctrl->next)
      {
      if (!ctrl->control->IsWindow ())
         ctrl->control->Create ();

      int h = ctrl->CtrlHeight ();
      if (h >= lineHeight)
         SetLineHeight (h);
      }
   if (lineHeight == 0)
      lineHeight = 16;  // default minimum height

   // column width ----------------------------------------------------------
   int x0 = 1;
	for (ctrl = firstColumn; ctrl; ctrl = ctrl->next)
		{
		int w = ctrl->CtrlWidth();
      ctrl->left = x0;
      ctrl->width = w+1;
      if (ctrl->type == gCheck)
         ctrl->width += 2;
      x0 += ctrl->width;

      // make the control a child of TGrid to clip it inside the grid -------
      if (ctrl->control)
         {
         ctrl->control->SetParent (this);
         ctrl->HideControl ();
         }
      }
   logicalWidth = lastColumn->left + lastColumn->width + 1;

   TRect voidRect;
   if (vscroll && vscroll->IsWindow())
      {
      vscroll->SetRange (0, maxLines < 0 ? nLines-1 : maxLines-1);
      vscroll->SetPosition (0);
      }

   if (hscroll && hscroll->IsWindow())
      {
      hscroll->LineMagnitude = LOWORD(GetDialogBaseUnits());// approx. 1 char
      hscroll->SetPosition (0);
      }

   if (frame)
      frame->SetWindowPos (HWND_BOTTOM, voidRect, SWP_NOMOVE|SWP_NOSIZE);

   // SetWindowPos will make sure everything is in place --------------------
   SetWindowPos (0, Attr.X, Attr.Y, Attr.W, Attr.H, SWP_NOZORDER);

   // Activate first column -------------------------------------------------
   if (delayFirstFocus)
      GotoFirstColumn ();
   }


void TGrid::SetHScrollRange ()
   {
   if (hscroll && hscroll->IsWindow())
      {
      hscroll->SetRange (0, logicalWidth - visibleWidth -1);
      hscroll->PageMagnitude = visibleWidth;
      }
   }

// ----------------------------------------------------------------------- //
// Data messages                                                           //
// ----------------------------------------------------------------------- //
bool TGrid::CanClose ()
	{
	int line = GetCurrentLine ();
   if (current)
      return (CanKillFocus (current) && ValidateLine (line));
   else
      return (ValidateLine (line));
	}


// ----------------------------------------------------------------------- //
// Mouse functions                                                         //
// ----------------------------------------------------------------------- //
bool TGrid::EvSetCursor (HWND hwnd, uint hitTest, uint mouseMsg)
   {
   if (hwnd == HWindow && hitTest == HTCLIENT)
      {
      TPoint p;
      GetCursorPos (p);
      if (SetCursorShape (p))
         return true;
      }

   return TControl::EvSetCursor (hwnd, hitTest, mouseMsg);
   }


bool TGrid::SetCursorShape (TPoint &p)
   {
   ScreenToClient (p);
   int x = p.x+1;

	int w = LOWORD(GetDialogBaseUnits());
   w += w;

   for (Column *e = firstColumn; e; e = e->next)
      {
      int right = e->left+e->width;
      if (x >= e->left && x < right)
         {
         TControl *c = e->control;
         switch (e->type)
            {
            default:
               SetCursor (0, IDC_ARROW);
               break;
            case gCombo:
               if (x < right-w &&
                  (c->Attr.Style & CBS_DROPDOWNLIST) != CBS_DROPDOWNLIST)
                  SetCursor (0, IDC_IBEAM);
               else
                  SetCursor (0, IDC_ARROW);
               break;
            case gDateEdit:
            case gTimeEdit:
            case gShortEdit:
            case gLongEdit:
            case gDoubleEdit:
            case gEdit:
               SetCursor (0, IDC_IBEAM);
               break;
            }
         return true;
         }
      }
   return false;
   }


void TGrid::EvLButtonDown (uint modKeys, TPoint &point)
   {
   ProcessMouse (modKeys, point, WM_LBUTTONDOWN);
   }


void TGrid::EvLButtonUp (uint modKeys, TPoint &point)
   {
   ProcessMouse (modKeys, point, WM_LBUTTONUP);
   }


void TGrid::EvLButtonDblClk (uint modKeys, TPoint &point)
   {
   ProcessMouse (modKeys, point, WM_LBUTTONDBLCLK);
   }


void TGrid::EvRButtonDown (uint modKeys, TPoint &point)
   {
   ProcessMouse (modKeys, point, WM_RBUTTONDOWN);
   }


void TGrid::EvRButtonUp (uint modKeys, TPoint &point)
   {
   ProcessMouse (modKeys, point, WM_RBUTTONUP);
   }


void TGrid::EvRButtonDblClk (uint modKeys, TPoint &point)
   {
   ProcessMouse (modKeys, point, WM_RBUTTONDBLCLK);
   }


void TGrid::ProcessMouse (uint modKeys, TPoint &point, uint message, bool sendToControl)
   {
   static bool echo = false;

   if (echo)
      {
      echo = false;
      return;
      }

   if (message == WM_LBUTTONDOWN)
      {
      int old = activeLine;
      int line = point.y / lineHeight;
      Column *e = FindColumn (point);
      ClientToScreen (point);

      // Validation ------------------------------------------------------------
      if (line != activeLine)
         {
         if (e != current)
            {
            if (!CanKillFocus (current))
               return;
            }

         // change line --------------------------------------------------------
         Go (firstLine+line, false);
         }

      if (e)
         {
         int ln = GetCurrentLine ();
         TControl *c = e->control;
         int id = c->GetId ();
         Visible v = VisibleCell (ln, id);
         if (e->type != gStatic && v != gEmpty && v != gLocked)
            {
            bool tabStop = ( (c->Attr.Style & WS_TABSTOP) ? true : false);
            if (dynamic_cast<TGCheckBox *>(c) && !tabStop)
               {
               bool val = *static_cast<const bool *>(TransferItem (ln, id));
               val ^= 1;
               if (DataChanging (ln, id, static_cast<void *>(&val)))
                  {
                  TransferItem (ln, id, static_cast<void *>(&val));
                  DataChanged (ln, id, static_cast<void *>(&val));
                  }
               RefreshItem (ln, id);
               }
            else
               {
               if (e != current || line != old)
                  e->SetFocus ();
               if (sendToControl)
                  {
                  echo = true;
                  c->ScreenToClient (point);
                  c->SendMessage (message, modKeys, MAKELONG(point.x,point.y));
                  }
               }
            }
         }

      if (staticOnly && lineSelection != SingleAutoSelect)
         SelectLine (GetCurrentLine ());
      }
   }


// ----------------------------------------------------------------------- //
// Borders and attributes                                                  //
// ----------------------------------------------------------------------- //
void TGrid::ChangeTitle (int id, const char *newTitle)
   {
   Column *e = FindColumn (id);
   if (e)
      e->SetTitle (newTitle);
   }

   
const char *TGrid::GetTitle (int id) const
   {
   Column *e = FindColumn (id);
   if (e)
      return e->title.c_str();
   else
      return 0;
   }


void TGrid::SetLineMarker (TGrid::LineMarker lm)
   {
   if (lm == lineMarker)
      return;

   // remove old markers ----------------------------------------------------
   if (vbar)
      vbar->ShowText (activeLine);

   lineMarker = lm;

   if (vbar)
      vbar->ShowMarker (activeLine);
   }


void TGrid::SetVBarText (int cwidth)
   {
   if (vbar)
      vbar->SetUserText (cwidth);
   }


void TGrid::SetFont (int id, const char *facename, int height, int weight,
                     bool italic)
   {
   if (!IsWindow())
      {
      Column *e = FindColumn (id);
      if (e)
         e->SetFont (facename, height, weight, italic);
      }
   }


void TGrid::SetFont (const char *facename, int height, int weight,
                     bool italic)
   {
   if (!IsWindow())
      {
      strcpy (gridLogFont->lfFaceName, facename);
      if (height)
         gridLogFont->lfHeight = height;
      gridLogFont->lfWeight = weight;
      gridLogFont->lfItalic = BYTE(italic);
      }
   }


void TGrid::SetTitleFont (const char *facename, int height, int weight,
                          bool italic)
   {
   if (!IsWindow ())
      {
      if (titleLogFont == 0)
         {
         titleLogFont = new LOGFONT;
         memcpy (titleLogFont, gridLogFont, sizeof(LOGFONT));
         }
      strcpy (titleLogFont->lfFaceName, facename);
      if (height)
         titleLogFont->lfHeight = height;
      titleLogFont->lfWeight = weight;
      titleLogFont->lfItalic = BYTE(italic);
      }
   }


void TGrid::SetFont ()
   {
   if (gridFont)
      delete gridFont;
   gridFont = new TFont (gridLogFont);

   // change fonts and redraw elements --------------------------------------
   WPARAM wp = reinterpret_cast<WPARAM>(HFONT(*gridFont));
   LPARAM lp = MAKELPARAM (true, 0);
   SendMessage (WM_SETFONT, wp, lp);

   for (Column *ctrl = firstColumn; ctrl; ctrl = ctrl->next)
      ctrl->SetFont ();
   }


void TGrid::SetTitleFont ()
   {
   if (titleFont)
      delete titleFont;
   if (titleLogFont == 0)
      {
      titleLogFont = new LOGFONT;
      memcpy (titleLogFont, gridLogFont, sizeof(LOGFONT));
      }
   titleFont = new TFont (titleLogFont);

   SetWindowFont (HFONT(*titleFont), true);
   if (vbar)
      vbar->SetWindowFont (HFONT(*titleFont), true);
   }


TColor TGrid::GetTitleColor (TGrid::TitleColor part) const
   {
   switch (part)
      {
      case TitleText:
         return GetSysColor (COLOR_BTNTEXT);
      case TitleBackground:
         return GetSysColor (COLOR_BTNFACE);
      case TitleHilite:
         return GetSysColor (COLOR_BTNHIGHLIGHT);
      case TitleShadow:
         return GetSysColor (COLOR_BTNSHADOW);
      case TitleFrame:
         return GetSysColor (COLOR_WINDOWFRAME);
      default:
         return TColor::Black;
      }
   }


TFont *TGrid::GetGridFont (int id) const
   {
   Column *e = FindColumn (id);
   if (e->font)
      return e->font;
   else
      return gridFont;
   }



void TGrid::SetVisibleCells (Visible v)
   {
   Visible old = visible;
   visible = v;
   if (old != v)
      Refresh ();
   }


void TGrid::HorzShrink ()
   {
   if (!IsWindow() || logicalWidth >= visibleWidth)
      return;

   int delta = visibleWidth - logicalWidth;
   SetWindowPos (0, 0,0, totalWidth-delta,totalHeight, SWP_NOZORDER|SWP_NOMOVE);
   }


int TGrid::GetTitleHeight () const
   {
   if (title)
      return title->CalcHeight ();
   else
      return 0;
   }


int TGrid::GetVBarWidth () const
   {
   if (vbar)
      return vbar->CalcWidth ();
   else
      return 0;
   }

// EOF
